import json
from datetime import date, datetime, time
from typing import Iterable

from .transformer import Transformer


class IsoJson(json.JSONEncoder):
    '''
    Encodes a dict to JSON with date, time and datetime expressed in ISO format.
    '''
    def _preprocess(self, obj):
        if isinstance(obj, (date, datetime, time)):
            return obj.isoformat()
        if isinstance(obj, dict):
            return {self._preprocess(k): self._preprocess(v) for k,v in obj.items()}
        if isinstance(obj, list):
            return [self._preprocess(i) for i in obj]
        return obj

    def default(self, obj):
        if isinstance(obj, (date, datetime, time)):
            return obj.isoformat()
        return super().default(obj)

    def iterencode(self, obj, _one_shot=False):
        return super().iterencode(self._preprocess(obj), _one_shot=_one_shot)
    
    @classmethod
    def dumps(cls, data):
        '''
        Returns a JSON string representation of data, 
        with conversion of date, time and datetime to 
        ISO format strings.
        '''        
        return cls().encode(data)

    @classmethod
    def dump_jsonl(cls, data):
        '''
        Returns a JSON-lines format representation of
        the list or data passed.
        '''
        instance = cls()

        if isinstance(data, Iterable):
            return '\n'.join((instance.dumps(item) for item in data))
        return instance.encode(data)


    """
    Attempts to parse any ISO date, time or datetime values to their native python types
    """
    parse_iso_dates = Transformer(time.fromisoformat, date.fromisoformat, datetime.fromisoformat)

    """
    Attempts to parse any date, time or datetime values to their native python types
    """
    parse_extra_dates = Transformer()\
        .enable_dates("%Y-%m-%d","%d/%m/%Y","%d-%b-%Y","%d-%m-%Y")\
        .enable_iso_dates()

    parse_with_timezone = Transformer(
        lambda txt : time.fromisoformat(txt), date.fromisoformat, datetime.fromisoformat)


    @classmethod
    def loads(cls, json_text : str, hook = None) -> dict:
        """
        Loads json data, with conversion of 
        ISO date, time and datetime to
        their respective Python types.
        """
        hook = hook or cls.parse_iso_dates
        return json.loads(json_text, object_hook=hook)


    @classmethod
    def load_lines(cls, json_text : str, hook = None) -> list:
        """
        Parses JSON-lines data and 
        returns a list of dicts, with 
        data parsed to native types using
        the given parser functions
        """
        hook = hook or cls.parse_iso_dates
        return [
            json.loads(txt, object_hook=hook) 
            for txt in json_text.splitlines()
            ] if json_text else []



from .typetwo import TypeTwo
from .rowkey import RowKey, NoKey, FunKey
from .dictcomparer import DictComparer
from .isojson import IsoJson
from .transformer import Transformer

__all__ = [
    TypeTwo.__name__, 
    RowKey.__name__, 
    NoKey.__name__, 
    FunKey.__name__,
    DictComparer.__name__, 
    IsoJson.__name__, 
    Transformer.__name__
]


from datetime import date, datetime
from pytz import utc
import types
from typing import Any, AnyStr, Callable, Dict, List, Tuple, Union
from .dictcomparer import DictComparer
from .rowkey import RowKey, FunKey, NoKey

class TypeTwo():
    """
    Performs Type-II slowly changed dimension data processing on a list of dicts
    """

    def __init__(self, 
        row_key     : Union[RowKey, str, List[str]],
        existing_rows : List[Dict[str, Any]] = [],
        from_field  : str  = '__FROM__', 
        to_field    : str  = '__TO__', 
        curr_field  : str   = '__CURRENT__', 
        from_date   : datetime = datetime.min,
        to_date     : datetime = datetime(2999,12,31), 
        fn_can_retire : Callable[[dict], bool] = lambda _ : True,
        ignore_fields = [],
        now         : datetime = datetime.now(utc)
    ):
        """
        Initializes an object which maintains state of a 
        set of rows upon which type-II slowly changed 
        dimension data processing is performed.

        existing_rows is a list of previously processed rows.
        row_key describes the primary key for records.

        from_field, to_field and current_key [all optional] are
        the names to be used for record validity fields. 
        i.e. fields to specify when a record became valid, 
        when it ceased to be valid, and whether it is 
        presently deemed valid.

        from_date and to_date are the default values to 
        assign to the record validity ranges.    
                
        fn_can_retire helps determine whether a record may be 
        retired, and defaults to always returning True. This 
        allows the caller to prevent the retirement of records
        that haven't appeared in the current source list.

        ignore_fields allows additional fields to be ignored
        when comparing rows for changes. This is useful when 
        extract timestamp values appear in source data.
        
        time_now defines the timestamp used for switching validity
        of any newly added record.

        now defines the datetime for when newly retired records end, 
        and their successor records start. 
        """
        if isinstance(row_key, str):
            self.row_key = RowKey(row_key)
        elif isinstance(row_key, list):
            self.row_key = RowKey(*row_key)
        elif row_key is None:
            self.row_key = NoKey()
        elif isinstance(row_key, types.FunctionType):
            self.row_key = FunKey(row_key)
        elif isinstance(row_key, RowKey):
            self.row_key = row_key
        else:
            raise NotImplementedError()

        self.from_field = from_field
        self.to_field = to_field
        self.current_key = curr_field
        self.from_date = from_date
        self.to_date = to_date 
        self.fn_can_retire = fn_can_retire 
        self.comparer = DictComparer(from_field, to_field, curr_field, *ignore_fields)
        self.visited_keys = set([])
        self.has_changes = False
        self.now_date = now

        for row in existing_rows:
            row.setdefault(from_field, from_date)
            row.setdefault(to_field, to_date)
            row.setdefault(curr_field, True)
  
        self.document = self.row_key.group_rows(existing_rows)
        
        for rows in self.document.values():
            if len(rows) > 1:                
                rows.sort(key = lambda r : r[from_field], reverse = True)
                for i, row in enumerate(rows):
                    if i > 0:
                        row[curr_field] = False

    def process_row(self, row, time_now = datetime.now()):
        """
        Args:
            row: dict   A now-current row from a system of record.
            self.time_now:   The timestamp to use for the change of record validity            

        Returns:
            True if the row was new or changed, else False if it had no effect on existing data.
        """
        
        if not isinstance(row, dict):
            raise NotImplementedError

        key = self.row_key(row)
        self.visited_keys.add(key)
        section = self.document.setdefault(key, [])
        existing = [e for e in section if e[self.current_key]]
        if existing:
            existing = existing[0]

            if self.comparer(row, existing):
                return False

            row.setdefault(self.from_field, time_now)
            existing[self.to_field] = row[self.from_field]
            existing[self.current_key] = False
        else:
            row.setdefault(self.from_field, self.from_date)

        row.setdefault(self.to_field, self.to_date)
        row.setdefault(self.current_key, True)        
        section.insert(0, row)
        self.has_changes = True
        return True


    def process_rows(self, current_rows : list):
        """
        Args:
            rows: list   The now-current rows from a system of record.
            
        Returns:
            A list of processed rows, sorted by section in descending order of change (i.e. newest first)
        """
        for row in current_rows:
            self.process_row(row)

        return iter(self)

    def __add__(self, other):
        """Add either a new row, or a list of new rows to the existing data"""
        if isinstance(other, list):
            self.process_rows(other)
        elif isinstance(other, dict):
            self.process_row(other)
        else:
            return NotImplemented
        return self            

    def __iter__(self):
        """Iterates over the SCD-II processed rows"""        
        for key, section in self.document.items():
            for i, row in enumerate(section):
                if i == 0:
                    if row[self.current_key] and (key not in self.visited_keys):
                        if self.fn_can_retire(row):
                            row[self.current_key] = False
                            row[self.to_field] = self.now_date
                            self.has_changes = True
                else:
                    row[self.current_key] = False
                    to_date = section[i-1][self.from_field] 
                    if row[self.to_field] is None or row[self.to_field] > to_date:
                        row[self.to_field] = to_date
                        self.has_changes = True
                yield row



from datetime import date, datetime
import types
from typing import Any, AnyStr, Callable, Dict, List, Tuple, Union
from .dictcomparer import DictComparer

class RowKey():
    """Encapsulates the fields used as the primary key for a record"""

    def __init__(self, *keys):
        """Encapsulates the fields used as the primary key for a record.
        ``keys``:  One of more key names (strings)"""
        self.keys = keys
    
    def __call__(self, row : dict) -> (Union[tuple,str]):
        """
        Returns a tuple of values from ``row`` which correspond to the given keys
        """
        
        if isinstance(row, dict):
            return tuple([row.get(key, None) for key in self.keys])
        raise NotImplementedError()
        
    def group_rows(self, rows : List[dict]):
        """Returns a dict where ``rows`` are grouped by this key"""
        document = {}
        for row in rows:
            document.setdefault(self(row), []).append(row)            
        return document
    
    def __iter__(self):
        """Iterates the key names"""
        return iter(self.keys)



class FunKey(RowKey):
    """Special case RowKey wraps a function"""
    def __init__(self, func_to_wrap):
        """Encapsulates the fields used as the primary key for a record."""
        self.func = func_to_wrap
    
    def __call__(self, row : dict):
        """
        Returns a tuple of values from ``row`` which correspond to the given keys
        """
        if isinstance(row, dict):
            return self.func(row)
            
        raise NotImplementedError()

    def __iter__(self):
        raise StopIteration()



class NoKey(RowKey):
    """Special case RowKey which groups all rows under a common key, namely 'ungrouped'"""
    def __init__(self):
        """Encapsulates the fields used as the primary key for a record."""
        self.keys = []

    def __call__(self, row : dict):
        """
        Returns a tuple of values from ``row`` which correspond to the given keys
        """
        return "ungrouped"



from datetime import date, datetime, time
from typing import Any, Callable
from pytz import timezone, utc

class Transformer():
    """
    A Transformer encapsulates a succession of
    functions which attempt to convert a value
    from one form to another.

    When called, the transformer applies each
    function in series, returning the value  
    from the first function to succesfully produce
    a new (different) value.

    ``in_place`` causes the transform to mutate values in 
    dicts and arrays when True. Otherwise, copies are 
    made and the original dict / array if left unchanged.

    Errors raised by transforms are ignored.
    """    
    def __init__(
        self, 
        *transforms
    ):
        self.transforms = list(transforms) if transforms else []


    def __add__(self, value : Callable[[Any], Any]):
        """Adds a transform function to this transformer.
        
        The transformer must be a simple function,
        taking a single value and returning a single value.
        
        Any errors raised by the transform function are ignored.
        """
        if isinstance(value, Transformer):
            return Transformer(*self.transforms, *value.transforms)
        if isinstance(value, function):            
            return Transformer(*self.transforms, value)
        else:
            raise NotImplementedError()

    def __call__(self, value, in_place : bool = True):
        """Applies this transform to the given ``value``. 
        Returns the original value if no transform succeeds"""

        if value is not None:  
            if isinstance(value, list):
                if in_place:
                    for i, d in enumerate(value):
                        value[i] = self(d, in_place = in_place) 
                    return value
                else:
                    return [self(d, in_place = in_place) for d in value]

            if isinstance(value, dict):
                if in_place:
                    for i, d in value.items():
                        value[i] = self(d, in_place = in_place) 
                    return value
                else:
                    return {key: self(v, in_place = in_place) for (key, v) in value.items()}
            
            for fn_xform in self.transforms:
                try:
                    new_value = fn_xform(value)
                    if new_value is not None and new_value != value:
                        return new_value
                except Exception as e:
                    pass
        return value
      
    def enable_dates(self, *formats):
        """Adds date format strings to recognize (see strptime)"""
        def make_formatter(format):
            return lambda d: datetime.strptime(d, format).date()
        self.transforms += [make_formatter(f) for f in formats]
        return self
    
    def enable_times(self, *formats):
        """Adds time format strings to recognize (see strptime)"""
        def make_formatter(format):
            return lambda d: datetime.strptime(d, format).time()
        self.transforms += [make_formatter(f) for f in formats]
        return self
    
    def enable_datetimes(self, *formats):
        """Adds datetime format strings to recognize (see strptime)"""
        def make_formatter(format):
            return lambda d: datetime.strptime(d, format)
        self.transforms += [make_formatter(f) for f in formats]
        return self

    def enable_iso_dates(self):
        """Adds support for parsing ISO date, time and datetime strings"""
        self.transforms += [time.fromisoformat, date.fromisoformat, datetime.fromisoformat]
        return self        

    def enable_datetimes_tz(self, source_tz : timezone = utc, target_tz : timezone = utc, *formats):
        """Adds support for parsing datetime with timezone correction.
                                
           source_tz: The assumed timezone of the source data being parsed.
           target_tz: The timezone which parsed dates should be represented in.
           formats: One or more strptime formats to parse. If no formats are provided, ISO format is inferred.
        """
        def make_formatter(format):            
            return lambda d: source_tz.localize(datetime.strptime(d, format)).astimezone(target_tz)
        
        def convert(dt):
            return source_tz.localize(datetime.fromisoformat(dt)).astimezone(target_tz)

        if formats:
            self.transforms += [make_formatter(f) for f in formats]  
        else:
            self.transforms += [convert]
        return self

